-- Toms Midi Mangler for Falcon :-)
-- Copyright 2015
-- Free for personal use, when using it in commercial libraries, I'd be happy if you send me a copy...  :-)
-- Contact me at thomas AT screendream DOT de

local modus = {
   {"Chromatic",      {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1}},
	{"Major",          {1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1}},
	{"Minor",          {1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 1, 0}},
	{"Harmonic Minor", {1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1}},
	{"Melodic Minor",  {1, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 1}},
   {"Pentatonic",     {1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0}},
	{"Dorian",         {1, 0, 1, 1, 0, 1, 0, 1, 0, 1, 1, 0}},
	{"Phrygian",       {1, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0}},
	{"Lydian",         {1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 1}},
	{"Mixolydian",     {1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0}},
	{"Locrian",        {1, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0}},
	{"Whole Tone",     {1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0}},
	{"Hungarian",      {1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 0}},
	{"Phrygish",       {1, 1, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0}},
   {"Blues",          {1, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1, 0}},
	{"Arabic",         {1, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 1}},
	{"Persian",        {1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1}},
	{"Acoustic",       {1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0}},
	{"Harmonic Major", {1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1}}
}

local scale_list = {}
for i,m in ipairs(modus) do
    table.insert(scale_list, m[1])
end

local note_list = {"C","C#","D","D#","E","F","F#","G","G#","A","A#","B"}

-- GUI definition:
NoteProb = Knob{"NoteProb", 76, 0, 100, true, 
                 position = {20, 10},
                 displayName = "Note Prob.",
                 tooltip = "Probability of the note to be played in percent."
                 }

VeloProb = Knob{"VeloProb", 76, 0, 100, true,
                 position = {160, 10},
                 displayName = "Velo. Prob.",
                 tooltip = "Probability for the note velocity to get changed randomly in percent. "
                 }

VeloDepth = Knob{"VeloDepth", 42, 0, 127, true,
                  position = {160, 60},
                  displayName = "Velo. Depth",
                  tooltip = "The maximum value of the random velocity change in Midi values from 0 - 127"
                  }

VeloDir = Menu{"VeloDirection", 
                {"Up", "Down", "Both"},
                position = {160, 110},
                value = 3,
                displayName = "Velocity Direction",
                tooltip = "Allows you to select if random velocity should only change upwards, only downwards or both"
                }

PanProb = Knob{"PanProb.", 75, 0, 100, true,
                position = {300, 10},
                displayName = "Pan Prob.",
                tooltip = "The probability of panning to be randomly changed around the center (per note)"
                }

PanDepth = Knob{"PanDepth", 50, 0, 100, true,
                 position = {300, 60},
                 displayName = "Pan Depth",
                 tooltip = "Maximum depth of the random panning change between 0 and 100 applied in both directions (per note)"
                 }

PitchProb = Knob{"PitchProb.", 26, 0, 100, true,
                  position = {430, 10},
                  displayName = "Pitch Prob.",
                  tooltip = "The probability of the note pitch to be changed randomly"
                  }

PitchDepth = Knob{"PitchDepth", 12, 0, 36, true,
                   position = {430, 60},
                   displayName = "Pitch Depth",
                   tooltip = "Maximum random note pitch change in halftones"                  
                  }

PitchDir = Menu{"PitchDirection", 
                {"Up", "Down", "Both"},
                position = {430, 110},
                value = 3,
                tooltip = "Direction of the random note pitch change: upwards, downwards or both"
               }

Scale = Menu{"Scale", 
              scale_list,
              position = {580, 10},
              value = 2,
              tooltip = "Select the scale you want the notes to conform to, otherwise select chromatic"
              }

Key = Menu{"Key", 
            note_list,
            position = {580, 60},
            value = 1,
            tooltip = "Select the key for your scale here..."
            }

Play = Menu{"Play", 
            {"Only Notes in Scale", "Transpose to Scale"},
            position = {580, 110},
            value = 2,
            tooltip = "Should notes outside the scale/key be removed or transposed to fit?"
            }

-- and this is where the magic happens :-)
function onNote(e)
   -- Note probability - will the note be played at all?
   if math.random(0, 99) < NoteProb.value then
      
      -- Velocity probability - will velocity be changed or not for the note
      local velo = e.velocity
      if math.random(0, 99) < VeloProb.value then
         -- If probability says yes, then we use a random value inside the set velocity depth range
         -- Only upwards
         if VeloDir.value == 1 then
            velo = velo + math.random(0, math.min(VeloDepth.value, 127 - velo))
         -- Only downwards
         elseif VeloDir.value == 2 then
            velo = velo - math.random(0, math.min(VeloDepth.value, velo))
         -- Up- and downwards
         elseif VeloDir.value == 3 then
            velo = velo + math.random(-math.min(VeloDepth.value, velo), math.min(VeloDepth.value, 127 - velo))
         end
      end
      
      -- Pan probability - should we change pan or not (per note)?
      local pan = 50
      if math.random(0, 99) < PanProb.value then
         -- Pan only has a simple random range around the center, should do in most cases I think... if not, let me know. :-)
         pan = pan + math.random(-PanDepth.value, PanDepth.value) * 0.01
      end

      -- Pitch probability - shall we change the note pitch?
      local pitch = e.note
      if math.random(0, 99) < PitchProb.value then
         -- Again we have up, down, or both on offer
         if PitchDir.value == 1 then
            pitch = pitch + math.random(0, math.min(PitchDepth.value, 127 - pitch))
         elseif PitchDir.value == 2 then
            pitch = pitch - math.random(0, math.min(PitchDepth.value, pitch))
         elseif PitchDir.value == 3 then
            pitch = pitch + math.random(-math.min(PitchDepth.value, pitch), math.min(PitchDepth.value, 127 - pitch))
         end
      end

      -- An finally we can put the result into a specific scale if we want
      -- First we extract one octave:
      local note = (pitch - Key.value + 1) % 12 + 1
      -- And retrieve the user selected scale pattern:
      local mode = modus[Scale.value][2]
      -- The first case is simple: Only notes that are in the scale pattern go through:
      if Play.value == 1 then
         if mode[note] == 1 then
            playNote(pitch, velo, -1, e.layer, e.channel, e.input, e.vol, pan)
         end
      -- The second case is more tricky
      elseif Play.value == 2 then
         -- First we let the notes through that fit the scale:
         if mode[note] == 1 then
            playNote(pitch, velo, -1, e.layer, e.channel, e.input, e.vol, pan)
         -- But now we have to care for the missfits:
         else
            local offset = 0
            -- A dummy variable that allows us to get a random value of either -1 or 1
            local move = {-1, 1}
            local tempNote = note
            -- We calculate a random offset, in steps of one up or down, since we don't want the direction always to be the same
            -- Since some scales have bigger gaps, we repeat this until we hit a scale note:
            repeat
               -- Here we calculate our offset. Since random returns values between and including the two integers, we can't directly
               -- get either -1 or 1 (one note up or one note down), so I create values of either 1 or 2 and use them as an index in the
               -- dummy variable "move" above. 
               offset = offset + move[math.random(1, 2)]
               -- Fron that we create our temporary note for checking it against the scale:
               tempNote = (pitch - Key.value + 1 + offset) % 12 + 1
               -- And we do that until we find a fit - not totally optimal maybe, since random can in theory create the same up and down
               -- movement many times before a fit is reached, but I would think the probability is low enough to not matter...
               -- Otherwise a fixed pattern could be used that only once is given a random starting direction.
            until mode[tempNote] == 1
            -- we apply the same offset to our original pitch:
            pitch = pitch + offset
            -- And finally send the newly calculated note out:
            playNote(pitch, velo, -1, e.layer, e.channel, e.input, e.vol, pan)
         end
      end
   end
end

function onRelease()
	-- eat event
end
